# Awesome intro to AI/ML

Hello there, and welcome traveler!

You've stumbled upon an inter-dimensional gap that will take you to a
fascinating world full of surprises, filled with animals such as Llamas,
Alpacas, Falcons and whatnot!

But fear not! I will be your guide in this daunting journey through the
space-time continuum.

## Prerequisites for inter-dimensional space travel (step 0)

You need to know Python! If that's not yet in your toolset, grab any Python
book / recommended Python course on the big net, and have at it. The guide's
recommendation is the [Harvard CS50
course](https://www.youtube.com/watch?v=OvKCESUCWII&list=PLhQjrBD2T3817j24-GogXmWqO5Q5vYy0V),
which is an introduction to programming in Python.

## The basics (Step 1)

If you've never heard of deep learning, natural language processing (NLP), large
language models (LLM), NumPy, PyTorch, then here is a great place to start! Even
if you know a bit about machine learning and artificial intelligence, I promise
you won't regret this!

First things first, let's get you accustomed to what machine learning is. This
course will get you up-to-date with the most basic stuff:
- [fast.ai course](https://course.fast.ai/): follow all of Part 1, don't go into
  Part 2. Don't be afraid to take your time, work the exercises in the course,
  and most importantly, don't be afraid to ask fellow astronauts around for
  help!

## The less basic (Step 2)

Now that you're suited up, you're prepared to jump through the inter-dimensional
gap!

At this point, you should be familiar with PyTorch, what deep learning is, and
how to implement a basic training loop in PyTorch using SGD (without using
fast.ai).

The next step is going a bit more in-depth into what PyTorch is all about, and
try to do a little deep-dive into the internals of how PyTorch works, and how to
implement architectural models without the use of the framework, manually
(that's how they did it in the old days):
- [Andrej Karpathy's Zero-to-Hero](https://karpathy.ai/zero-to-hero.html):
  follow all the tutorials, WITHOUT the last one (we'll get back to it later).
  This will teach you how to fix your spaceship if it ever breaks, and how to
  hike in space if you ever need to!

## The soon-to-be-intermediate (step 3)

Now, we're getting to the meat of things. You've jumped through the portal, and
you see a completely new world, full of fascination. You scream "I want
MOOOORE", but no one is here to help you. Everyone's already left the landing
place. They've built modern space shuttles, and gone to the closest restaurant
to grab a space burger. It's time you do the same!

Now, here comes the fascinating stuff. Everything after 2017 revolves around a
singular architecture which transformed everything in NLP. That's right, it's
the TRANSFORMER (*thunder bolt in the background*). You need to get familiar
with this one if you want to build your own modern space shuttle:
- ["Attention Is All You Need" (Vaswani et al.,
  2017)](https://arxiv.org/abs/1706.03762): this is the original paper of the
  architecture, which you need to understand well in order to progress further
  in the field of NLP. Read it once. Admire the fact that you don't understand
  anything. Now it's time to do some digging and understand it. Trust me, it's
  important. Here are a few resources to help you out: [1][1], [2][2], [3][3].
  For anything else you don't understand, you need to search in the hitchhiker's
  guide to the galaxy, or the closest astronaut able to help you.

[1]: https://jalammar.github.io/illustrated-transformer/
[2]: https://towardsdatascience.com/all-you-need-to-know-about-attention-and-transformers-in-depth-understanding-part-1-552f0b41d021
[3]: https://towardsdatascience.com/transformer-models-101-getting-started-part-1-b3a77ccfa14d

After understanding the paper, you need to implement it! (*duuuh*) Now it's time
to go back to Andrej Karpathy's course (listed above) and do that last one
(titled "Let's build GPT" at the time of writing).

## The intermediates (step 4)

Well, you've learnt how to get yourself suited, jump in an inter-dimensional
gap, land on the other side without being hurt, learned how to ride *AND* build
a modern space shuttle. Now it's time to start that engine and get on your way
to that space burger!

At this point, you should be familiar with how a Transformer works, how to
implement one, how to train one, and, hopefully, how to read a paper related to
such a model.

The next step is to start getting closer to the present! Currently, NetSys works
on a model released by Meta Research in july 2023, called LLaMA 2, open-sourced
to the public. Read (and understand) the 2 papers for the original LLaMA and the
spin-off. Keep in mind, there has been 6 years since the original Transformer,
so there have been a number of modifications to the architecture in the
meanwhile. Anything you don't understand, you need to go to the references of
the paper, and read about that thing you don't understand from the other paper.
Rinse and repeat, however deep or wide you need to go:
- [LLaMA: Open and Efficient Foundation Language Models](https://arxiv.org/abs/2302.13971)
- [Llama 2: Open Foundation and Fine-Tuned Chat Models](https://arxiv.org/abs/2307.09288)

After reading the papers, go read the source code and try to run the models
(it's not very complicated, just a few lines of PyTorch code you know so well
already).

## The advanced (step 5)

Congratulations! You've now reached the space hub where astronauts meet to talk
space things and eat space burgers!

At this point, you understand how a Transformer works, what the current
state-of-the-art is in terms of models, text generation, weights etc.

In a moment you weren't looking, the bartender at the space restaurant hands you
out a map, and starts talking space gibberish to you. You don't understand a
thing, but there seems to be a few things you can figure out from the map. Ah!
He's trying to hire you for a space adventure, however it seems that he lets you
choose whatever you want to do! You take a closer look at the map, and you
recognize, among the dozen or so points of interest, a few which sound familiar:
multi-GPU parallelization, datacenter communication patterns, weights
quantization, memory/transfer profiling, code generation, formal verification,
model training and fine-tuning, prompt engineering.

Well, traveler, it's time for you now to choose your adventure!

*(Head over to [the astronaut's second intro to timely research
papers](doc/intro2.md) (some of these you might've - and should've - already
read while you were at it from the LLaMA papers above), or [the research
doc](doc/llm.md) to see what other astronauts have been up to)*
